from functools import wraps
from typing import Any, Callable, List

import regex


def substitute_brackets(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(self: Any, string: str, *args: Any, **kwargs: Any) -> List[Any]:
        reg = r"(\((?>[^()]+|(?1))*\))"
        counter = -1

        def count() -> int:
            nonlocal counter
            counter += 1
            return counter

        # Substitute brackets
        brackets = regex.findall(reg, string)
        substituted = regex.sub(reg, lambda x: "{" + str(count()) + "}", string)

        # Use substitution in function
        result = func(self, substituted, *args, **kwargs)

        # Allow robust indexing
        if isinstance(result, tuple):
            result = list(result)
        else:
            result = [result]

        # Reapply brackets (assume last variable needs substitution)
        if isinstance(result[-1], str):
            result[-1] = result[-1].format(*brackets)
        elif isinstance(result[-1], list):
            result[-1] = [string.format(*brackets) for string in result[-1]]
        elif isinstance(result[-1], dict):
            result[-1] = {key: value.format(*brackets) for key, value in result.items()}

        if len(result) == 1:
            return result[0]
        return result

    return wrapper



import http.server
import re
import shutil
import socketserver
import webbrowser
from os import fspath, listdir, path
from pathlib import Path
from typing import Any, List, Set, Tuple, Union
from uuid import uuid1

import networkx as nx
import pydot
from graphviz import Digraph

from .parser import SqlParser


class LineageGraph:
    def __init__(self):
        self.parser = SqlParser()
        self._node_attrs = {"color": "lightblue2", "style": "filled", "shape": "rect"}
        self._clause_attrs = {"shape": "diamond", "color": "lightgrey"}
        self._source_attrs = {"color": "palegreen"}
        self._inner_query_attrs = {"shape": "ellipse", "color": "lightgoldenrod2"}
        self._union_query_attrs = {"shape": "ellipse", "color": "lightcoral"}

    def _get_sources(self, parsed: Union[list, dict]) -> List[Any]:
        "Split parsed into sources. Will be iterated over to build edges."
        if isinstance(parsed, dict):
            # End-source or inner query
            return [
                (clause, value["content"])
                for clause, value in parsed.items()
                if "join" in clause or clause == "from"
            ]
        else:
            # Union query
            return [(None, parsed)]

    def _get_edges(
        self, target: str, parsed: dict, edges: set
    ) -> Set[Tuple[str, str, str]]:
        "Get all edges for lineage graph to be extended."

        # Split parsed into standardized sources
        sources = self._get_sources(parsed)

        # Iterate over sources, handle three options differently
        for clause, source in sources:

            # 1. End target
            if isinstance(source, str):
                edges.add((target, clause, source))
            # 2. Inner query
            elif isinstance(source, dict):
                inner_query = f"Inner Query {uuid1()}"
                edges.add((target, clause, inner_query))
                edges.update(self._get_edges(inner_query, source, edges))
            # 3. Union query
            elif isinstance(source, list):
                union_query = f"Union Query {uuid1()}"
                edges.add((target, None, union_query))

                for union_source in source:
                    inner_query = f"Inner Query {uuid1()}"
                    edges.add((union_query, None, inner_query))
                    edges.update(self._get_edges(inner_query, union_source, edges))

        return edges

    def _extend_graph(self, top_level_target: str, parsed: dict) -> None:
        """Extend graph by collecting edges and adding conditionally based
        on source type."""
        edges = self._get_edges(top_level_target, parsed, edges=set())

        if not hasattr(self, "edges"):
            self.edges = set()
        self.edges.update(edges)

        for target, clause, source in edges:
            # Standardize label for inner/union queries
            if "Inner Query" in target:
                self.graph.node(target, label="Inner Query", **self._inner_query_attrs)
            if "Union Query" in target:
                self.graph.node(target, label="Union Query", **self._union_query_attrs)

            if clause:
                # Unique connector with clause and target
                connector = f"Connector-{clause}-{target}"
                self.graph.node(connector, label=clause, **self._clause_attrs)
                # Add final edges, through connector
                self.graph.edge(source, connector)
                self.graph.edge(connector, target)
            else:
                self.graph.edge(source, target)

        targets, clauses, sources = zip(*self.edges)

        nodes = set(sources) | set(targets)
        source_nodes = set(sources) - set(targets)

        for node in nodes:
            if not node.startswith(("Connector", "Inner Query", "Union Query")):
                self.graph.node(
                    node,
                    **(
                        self._source_attrs if node in source_nodes else self._node_attrs
                    ),
                )

    def _get_top_level_target(self, query: str) -> str:
        "Find target: Table or View to be created."
        reg = r"^\s*(?:CREATE\s+(?:OR REPLACE\s+)?(?:VIEW|TABLE)\s+)([\S]+)"
        result = re.findall(reg, query, flags=re.I)
        if len(result) == 0:
            raise ValueError(
                "Please provide input string with a CREATE TABLE/VIEW statement."
            )
        target = result[0]
        return target

    def instantiate_fresh_graph(self) -> None:
        self.graph = Digraph(
            "Lineage Graph",
            # Standard node attrs for sources
            node_attr=self._node_attrs,
            # Disallow duplication
            strict=True,
        )

    def extend_graph_from_input_string(self, input_string: str) -> None:
        "Provide a create statement to extend the data lineage graph."
        if not hasattr(self, "graph"):
            self.instantiate_fresh_graph()

        top_level_target = self._get_top_level_target(input_string)
        parsed = self.parser.parse(input_string)
        self._extend_graph(top_level_target, parsed)

    def create_graph_from_directory(self, dir: str) -> None:
        "Build data lineage graph from data directory."
        self.instantiate_fresh_graph()

        files = listdir(dir)
        sql_files = filter(lambda f: f.endswith(".sql"), files)

        for file_name in sql_files:
            with open(path.join(dir, file_name), "r") as f:
                input_string = f.read()
                self.extend_graph_from_input_string(input_string)

    def get_execution_order(self) -> List[str]:
        "Get execution order of tables / views."
        if not hasattr(self, "graph"):
            raise ValueError("No graph has been created yet")

        dot_graph = pydot.graph_from_dot_data(self.graph.source)[0]
        nx_graph = nx.nx_pydot.from_pydot(dot_graph)

        execution_order = [
            node
            for node in nx.topological_sort(nx_graph)
            if not node.startswith(("Connector", "Inner Query", "Union Query"))
        ]

        return execution_order

    def serve_graph(self) -> None:
        "Serve graph as .html file"

        if not hasattr(self, "graph"):
            raise ValueError("No graph has been created yet")

        PORT = 8000
        DIRECTORY = "out"

        print(f"Writing files to /{DIRECTORY}")

        # Render graph as .gv and .gv.pdf
        self.graph.render(directory=DIRECTORY)

        # Copy template index.html to load .gv using d3-graphviz
        PACKAGE_PATH = fspath(Path(__file__).parent)
        shutil.copy(
            path.join(PACKAGE_PATH, "templates", "index.html"),
            f"{DIRECTORY}/index.html",
        )

        def start_server() -> None:
            class Handler(http.server.SimpleHTTPRequestHandler):
                def __init__(self, *args: Any, **kwargs: Any):
                    super().__init__(
                        directory=DIRECTORY, *args, **kwargs
                    )  # type: ignore

            with socketserver.TCPServer(("", PORT), Handler) as httpd:
                print(f"Serving at port {PORT}")
                httpd.serve_forever()

        def open_webbrowser() -> None:
            webbrowser.open(f"http://localhost:{PORT}", new=2)

        open_webbrowser()
        start_server()



import re
from typing import Any, Dict, List, Tuple, Union

from .helpers import substitute_brackets


class SqlParser:
    clauses = [
        "SELECT",
        "FROM",
        "LEFT JOIN",
        "RIGHT JOIN",
        "FULL JOIN",
        "FULL OUTER JOIN",
        "INNER JOIN",
        "WHERE",
        "QUALIFY",
    ]

    @classmethod
    def _build_clause_regex(cls) -> str:
        """Builds main regex to capture all clauses, as identified above."""
        reg_base = (r"(?:^|\s+){}\s+|" * len(cls.clauses))[:-1]
        reg = reg_base.format(*cls.clauses)
        return reg

    def __init__(self):
        pass

    def _clean_string(self, string: str) -> str:
        string = re.sub(r"#.*", "", string)  # remove comments
        string = re.sub(r"\s+", " ", string)  # standardize whitespace
        string = re.sub(r"\/\*.*\*\/", "", string)  # remove comments
        string = string.strip().lower()
        return string

    def _clean_list(self, lst: List[str]) -> List[str]:
        cleaned = filter(lambda item: item, lst)  # remove empty values
        return list(map(self._clean_string, cleaned))

    @substitute_brackets
    def _clean_split(
        self, string: str, pattern: str = r"\s+", maxsplit: int = 0
    ) -> List[str]:
        return self._clean_list(re.split(pattern, string, maxsplit, flags=re.I))

    def _split_value_into_content_and_alias(self, value: str) -> Dict[str, str]:
        """Function for aliasing (used in source clauses and SELECT items)"""
        keys = ["content", "alias"]
        return dict(zip(keys, self._clean_split(value, r"\s+as\s+")))

    def _split_select_value(self, value: str) -> List[Dict[str, str]]:
        """SELECT statement value splitting"""
        select_statements = self._clean_split(value, ",")
        return list(map(self._split_value_into_content_and_alias, select_statements))

    def _split_source_value(self, value: str) -> Dict[str, str]:
        """Value splitting for sources (FROM / JOIN)"""
        split_on_whitespace = self._clean_split(value)

        if len(split_on_whitespace) == 2:
            split_on_whitespace.insert(1, "as")

        value = " ".join(split_on_whitespace)
        return self._split_value_into_content_and_alias(value)

    def _split_join_on(self, value: str) -> Dict[str, List[str]]:
        """JOIN statement value splitting"""
        return {"on": self._clean_split(value, r"\s+and\s+")}

    def _split_join_value(self, value: str) -> Dict[str, Union[str, List[str]]]:
        """Join on value splitting"""
        source, join_on = self._clean_split(value, r"\s+on\s+", maxsplit=1)

        return {**self._split_source_value(source), **self._split_join_on(join_on)}

    def _split_condition_value(self, value: str) -> List[str]:
        """Condition statement value splitting (WHERE / QUALIFY)"""
        return self._clean_split(value, r"\s+and\s+")

    def _split_value(self, clause: str, value: str) -> Any:
        """Function for splitting values based on right clause"""
        return (
            self._split_select_value(value)
            if clause == "select"
            else self._split_source_value(value)
            if clause == "from"
            else self._split_join_value(value)
            if "join" in clause
            else self._split_condition_value(value)
            if clause in ("where", "qualify")
            else value
        )

    @substitute_brackets
    def _split_clauses_and_values(self, query: str) -> Tuple[List[str], List[str]]:
        """Starting point for parsing. Split into clauses and values."""
        reg = self._build_clause_regex()

        clauses = self._clean_list(re.findall(reg, query, flags=re.I))
        values = self._clean_list(re.split(reg, query, flags=re.I))

        return clauses, values

    def parse(self, input_string: str) -> Union[list, dict]:
        """Callable for user."""
        return self._parse(self._clean_string(input_string))

    def _parse(self, input_string: str) -> Union[list, dict]:
        """Test for unions -> leads to different parsing layout."""
        union_split = self._clean_split(input_string, r"UNION\s+(?:ALL)?")

        if len(union_split) == 1:
            return self._parse_query(input_string)
        else:
            return [self._parse_query(query) for query in union_split]

    def _parse_query(self, query: str) -> dict:
        """Parse a complete query (assume starts with SELECT)"""
        clauses, values = self._split_clauses_and_values(query)

        # If clause is not the first argument, omit what precedes
        # first encountered clause
        values = values if len(clauses) == len(values) else values[1:]

        combinations = {}
        for clause, value in zip(clauses, values):
            split_value = self._split_value(clause, value)

            # Detect inner SQL
            if "join" in clause or clause == "from":
                if "select" in split_value["content"]:
                    without_brackets = (
                        split_value["content"].strip().lstrip("(").rstrip(")")
                    )
                    # Continue parsing inner SQL
                    split_value["content"] = self._parse(without_brackets)

            # Re-join clause and value
            combinations[clause] = split_value

        return combinations



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pylineage']

package_data = \
{'': ['*'], 'pylineage': ['templates/*']}

install_requires = \
['graphviz>=0.16,<0.17',
 'networkx>=2.5.1,<3.0.0',
 'pydot>=1.4.2,<2.0.0',
 'regex>=2021.4.4,<2022.0.0']

setup_kwargs = {
    'name': 'pylineage',
    'version': '0.1.0',
    'description': 'Data Lineage for Python',
    'long_description': '# pylineage\nData Lineage for Python\n',
    'author': 'jasperpaalman',
    'author_email': 'jasper_paalman@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jasperpaalman/pylineage',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
